from pathlib import Path
from typing import Annotated

import typer
from loguru import logger
from markdown import markdown
from weasyprint import HTML

from smart_letters.exceptions import handle_abort, Abort
from smart_letters.cache import init_cache, BACKUP_DIR
from smart_letters.config import attach_settings
from smart_letters.format import simple_message
from smart_letters.schemas import RenderConfig
from smart_letters.utilities import asset_path


cli = typer.Typer()


@cli.callback(invoke_without_command=True)
@handle_abort
@init_cache
@attach_settings
def render(
    ctx: typer.Context,
    md_file: Annotated[Path, typer.Argument(help="The path to the markdown file to render.")],
    file_stem: Annotated[str | None, typer.Option(help="The stem of the file to save.")] = None,
):
    """
    Render a Markdown cover letter to styled PDF.

    This command is useful if you just want to render a Markdown letter without generating one.
    This might be useful, for example, if you are using a cached letter.
    """
    render_config = RenderConfig(
        file_stem=file_stem or md_file.stem,
        timestamp=ctx.obj.timestamp,
        output_directory=ctx.obj.settings.output_directory,
    )

    try:
        letter_text = md_file.read_text()
    except Exception:
        raise Abort(f"Couldn't read the markdown file from {md_file}", subject="Render failed")

    pdf_path = render_letter(letter_text, render_config)
    simple_message(f"Letter rendered to {pdf_path}")


def render_letter(
    letter_text: str,
    render_config: RenderConfig,
) -> Path:
    logger.debug("Rendering markdown to file")

    pdf_path = Path(f"{render_config.file_stem}.pdf")
    if render_config.output_directory:
        pdf_path = render_config.output_directory / pdf_path

    html_content = markdown(letter_text)
    html_file_path = BACKUP_DIR / Path(f".{render_config.timestamp}.{render_config.file_stem}.html")
    logger.debug(f"Dumping html to {html_file_path}")
    html_file_path.write_text(html_content)

    css_paths = [asset_path("styles.css")]
    html = HTML(string=html_content)
    html.write_pdf(pdf_path, stylesheets=css_paths)

    logger.debug(f"Letter rendered to {pdf_path}")
    return pdf_path



from dataclasses import dataclass
import json
import subprocess
from pathlib import Path
from typing import Annotated

import httpx
import typer
import pyperclip
from inflection import parameterize
from loguru import logger
from openai import OpenAI

from smart_letters.cache import BACKUP_DIR, init_cache
from smart_letters.config import attach_settings
from smart_letters.exceptions import handle_abort
from smart_letters.format import terminal_message, simple_message
from smart_letters.prompt import get_prompts, build_prompts
from smart_letters.utilities import asset_path
from smart_letters.render import render_letter
from smart_letters.schemas import LetterConfig, PromptConfig, RenderConfig


@dataclass
class Reprompt:
    old_letter: str
    user_feedback: str


def pull_posting(letter_config: LetterConfig) -> str:
    logger.debug(f"Pulling posting from {letter_config.posting_url}")
    response = httpx.get(letter_config.posting_url)
    response.raise_for_status()
    return response.text


def pull_resume(letter_config: LetterConfig) -> str:
    logger.debug(f"Loading resume from {letter_config.resume_path}")
    return letter_config.resume_path.read_text()


def pull_example(letter_config: LetterConfig) -> str | None:
    if letter_config.example_path:
        logger.debug(f"Loading example text from {letter_config.example_path}")
        return letter_config.example_path.read_text()
    return None


def pull_heading(letter_config: LetterConfig) -> str | None:
    if letter_config.heading_path:
        logger.debug(f"Loading heading text from {letter_config.heading_path}")
        return letter_config.heading_path.read_text()
    return None


def generate_letter(
    prompt_config: PromptConfig,
    letter_config: LetterConfig,
    reprompts: list[Reprompt],
    posting_text: str,
    resume_text: str,
    example_text: str | None = None,
) -> str:
    logger.debug("Generating letter using OpenAI")
    client = OpenAI(api_key=letter_config.openai_api_key)

    prompts = get_prompts(prompt_config)
    messages = build_prompts(
        prompts,
        posting_text=posting_text,
        resume_text=resume_text,
        example_text=example_text,
    )

    for reprompt in reprompts:
        messages.append(dict(role="assistant", content=reprompt.old_letter))
        messages.append(dict(role="user", content=reprompt.user_feedback))

    kwargs = dict(model="gpt-4o", n=1)
    logger.debug(f"Using params for OpenAI: \n{json.dumps(kwargs, indent=2)}")
    cmp = client.chat.completions.create(messages=messages, **kwargs)  # type: ignore
    text = cmp.choices[0].message.content
    assert text is not None
    return text


def assemble_letter(
    letter_config: LetterConfig,
    letter_text: str,
    heading_text: str | None,
) -> str:
    logger.debug("Assembling letter")
    parts = []

    if heading_text:
        parts.append(heading_text)

    if letter_config.company:
        parts.append(f"To the Hiring Team at {letter_config.company}:")
    else:
        parts.append("Dear Hiring Manager,")

    parts.append("")
    parts.append(letter_text)
    parts.append("")
    parts.append("Best regards,")
    parts.append("")

    if letter_config.sig_path:
        parts.append(f"![signature](file://{letter_config.sig_path.absolute()})")
        parts.append("")

    parts.append(letter_config.candidate_name)

    return "\n".join(parts)


def show_letter(letter_text: str, footer_text: str | None = None):
    terminal_message(letter_text, subject="Generated Letter", markdown=True, footer=footer_text)


def request_feedback(letter_text: str) -> str:
    show_letter(letter_text)
    accepted = typer.confirm("Are you satisfied with the letter?", default=True)
    if accepted:
        return ""
    return typer.prompt("What can I do to fix it?", default="Just try again")


def should_render(cli_param: bool | None) -> bool:
    if cli_param is not None:
        return cli_param
    return typer.confirm("Should I render the letter to PDF?", default=True)


def read_cached_letter(letter_config: LetterConfig) -> str:
    return letter_config.cache_path.read_text()


def write_cached_letter(letter_config: LetterConfig, letter_text):
    logger.debug(f"Writing current version of letter to {letter_config.cache_path}")
    letter_config.cache_path.write_text(letter_text)


def edit_letter(letter_config: LetterConfig, letter_text: str):
    edit = typer.confirm("Would you like to edit the letter?", default=False)
    if not edit:
        return letter_text

    write_cached_letter(letter_config, letter_text)
    logger.debug(f"Editing generated letter at {letter_config.cache_path}")
    subprocess.run([letter_config.editor_command, str(letter_config.cache_path)])
    return read_cached_letter(letter_config)


def get_file_stem(letter_config: LetterConfig) -> str:
    stem = letter_config.filename_prefix
    if letter_config.company:
        stem += f"--{parameterize(letter_config.company)}"
    if letter_config.position:
        stem += f"--{parameterize(letter_config.position)}"
    return stem


def build_letter(prompt_config: PromptConfig, letter_config: LetterConfig) -> str:
    logger.debug("Building letter")
    full_text = ""
    heading_text: str | None = pull_heading(letter_config)
    if letter_config.fake:
        logger.debug("Using fake text for letter")
        letter_text = asset_path("fake.txt").read_text()
        full_text = assemble_letter(letter_config, letter_text, heading_text)
        write_cached_letter(letter_config, full_text)
    else:
        logger.debug("Generating letter")
        posting_text: str = pull_posting(letter_config)
        resume_text: str = pull_resume(letter_config)
        example_text: str | None = pull_example(letter_config)

        accepted = False
        reprompts: list[Reprompt] = []

        while not accepted:
            letter_text = generate_letter(
                prompt_config, letter_config, reprompts, posting_text, resume_text, example_text
            )
            full_text = assemble_letter(letter_config, letter_text, heading_text)
            write_cached_letter(letter_config, full_text)
            feedback = request_feedback(full_text)
            if feedback:
                logger.debug("Letter was not accepted, requesting feedback")
                reprompts.append(Reprompt(old_letter=full_text, user_feedback=feedback))
            else:
                accepted = True

    full_text = edit_letter(letter_config, full_text)

    return full_text


cli = typer.Typer()


@cli.callback(invoke_without_command=True)
@handle_abort
@init_cache
@attach_settings
def generate(
    ctx: typer.Context,
    posting_url: Annotated[str, typer.Argument(help="The URL of the job posting.")],
    company: Annotated[str | None, typer.Option(help="The name of the company.")] = None,
    position: Annotated[str | None, typer.Option(help="The title for the job.")] = None,
    example_letter: Annotated[Path | None, typer.Option(help="An example letter to use as a reference.")] = None,
    fake: Annotated[
        bool,
        typer.Option(help="[FOR DEBUGGING] Use a fake letter body instead of fetching one from OpenAI."),
    ] = False,
    render: Annotated[
        bool | None,
        typer.Option(
            help="""
                If this option is provided, render (or don't render) the generated Markdown to PDF.
                If not provided, you will be asked if you want to render it after the letter is generated.
            """,
        ),
    ] = None,
):
    """
    Generate a cover letter for a job posting listed at the provided URL.

    This command will use an LLM prompt (either builtin or user provided in config) to generate
    a cover letter (in Markdown). You may also render the letter into a styled PDF.
    """
    prompt_config = PromptConfig(
        dev_prompt_path=ctx.obj.settings.dev_prompt_path,
        user_prompt_template_path=ctx.obj.settings.user_prompt_template_path,
    )

    letter_config = LetterConfig(
        resume_path=ctx.obj.settings.resume_path,
        candidate_name=ctx.obj.settings.candidate_name,
        filename_prefix=ctx.obj.settings.filename_prefix,
        openai_api_key=ctx.obj.settings.openai_api_key,
        cache_path=BACKUP_DIR.joinpath(f".{ctx.obj.timestamp}.md"),
        editor_command=ctx.obj.settings.editor_command,
        sig_path=ctx.obj.settings.sig_path,
        heading_path=ctx.obj.settings.heading_path,
        example_path=example_letter,
        output_directory=ctx.obj.settings.output_directory,
        company=company,
        position=position,
        posting_url=posting_url,
        fake=fake,
    )

    letter_text: str = build_letter(prompt_config, letter_config)

    if should_render(render):
        render_config = RenderConfig(
            file_stem=get_file_stem(letter_config),
            timestamp=ctx.obj.timestamp,
            output_directory=ctx.obj.settings.output_directory,
        )
        pdf_path = render_letter(letter_text, render_config)
        simple_message(f"Letter saved to {pdf_path}")
    else:
        show_kwargs = {}
        try:
            pyperclip.copy(letter_text)
            show_kwargs["footer_text"] = "Letter copied to clipboard"
        except Exception as exc:
            logger.debug(f"Could not copy letter to clipboard: {exc}")
        show_letter(letter_text, **show_kwargs)



from contextlib import contextmanager
import json
from functools import wraps
from pathlib import Path
from typing import Annotated, Any
import os

import snick
import typer
from inflection import dasherize
from loguru import logger
from pydantic import AfterValidator, BaseModel, ValidationError, Field

from smart_letters.exceptions import Abort, handle_abort
from smart_letters.cache import CACHE_DIR, init_cache
from smart_letters.format import terminal_message


settings_path: Path = CACHE_DIR / "settings.json"


def file_exists(value: Path | None) -> Path | None:
    if value is None:
        return value

    value = value.expanduser()
    if not value.exists():
        raise ValueError(f"File not found at {value}")
    return value


def has_editor(value: str | None) -> str:
    if value is not None:
        return value

    try:
        return os.environ["EDITOR"]
    except Exception:
        raise ValueError("Couldn't load editor from environment. Please set it explicitly")


class Settings(BaseModel):
    openai_api_key: str
    resume_path: Annotated[Path, AfterValidator(file_exists)]
    candidate_name: str
    filename_prefix: str = "cover-letter"
    heading_path: Annotated[Path | None, AfterValidator(file_exists)] = None
    sig_path: Annotated[Path | None, AfterValidator(file_exists)] = None
    output_directory: Annotated[Path | None, AfterValidator(file_exists)] = None
    dev_prompt_path: Annotated[Path | None, AfterValidator(file_exists)] = None
    user_prompt_template_path: Annotated[Path | None, AfterValidator(file_exists)] = None
    editor_command: Annotated[str | None, AfterValidator(has_editor)] = None

    invalid_warning: Annotated[
        str | None,
        Field(
            exclude=True,
            description="""
            An optional warning that can be included when the model is invalid.

            Used when we use the `attach_settings` decorator with `validate=False`.
        """,
        ),
    ] = None


@contextmanager
def handle_config_error():
    try:
        yield
    except ValidationError as err:
        raise Abort(
            snick.conjoin(
                "A configuration error was detected.",
                "",
                "Details:",
                "",
                f"[red]{err}[/red]",
            ),
            subject="Configuration Error",
            log_message="Configuration error",
        )


def init_settings(validate: bool = True, **settings_values) -> Settings:
    with handle_config_error():
        logger.debug("Validating settings")
        try:
            return Settings(**settings_values)
        except ValidationError as err:
            if validate:
                raise
            settings = Settings.model_construct(**settings_values)
            settings.invalid_warning = str(err)
            return settings


def update_settings(settings: Settings, **settings_values) -> Settings:
    with handle_config_error():
        logger.debug("Validating settings")
        settings_dict = settings.model_dump(exclude_unset=True)
        settings_dict.update(**settings_values)
        return Settings(**settings_dict)


def unset_settings(settings: Settings, *unset_keys) -> Settings:
    with handle_config_error():
        logger.debug("Unsetting settings")
        return Settings(**{k: v for (k, v) in settings.model_dump(exclude_unset=True).items() if k not in unset_keys})


def attach_settings(original_function=None, *, validate=True):
    """
    Attach the settings to the CLI context.

    Optionally, skip validation of the settings. This is useful in case the config
    file being loaded is not valid, but we still want to use the settings. Then, we
    can update the settings with correct values.

    Uses recipe for decorator with optional arguments from:
    https://stackoverflow.com/a/24617244/642511
    """

    def _decorate(func):
        @wraps(func)
        def wrapper(ctx: typer.Context, *args, **kwargs):
            try:
                logger.debug(f"Loading settings from {settings_path}")
                settings_values = json.loads(settings_path.read_text())
            except FileNotFoundError:
                raise Abort(
                    f"""
                    No settings file found at {settings_path}!

                    Run the set-config sub-command first to establish your settings.
                    """,
                    subject="Settings file missing!",
                    log_message="Settings file missing!",
                )
            logger.debug("Binding settings to CLI context")
            ctx.obj.settings = init_settings(validate=validate, **settings_values)
            return func(ctx, *args, **kwargs)

        return wrapper

    if original_function:
        return _decorate(original_function)
    else:
        return _decorate


def dump_settings(settings: Settings):
    logger.debug(f"Saving settings to {settings_path}")
    settings_values = settings.model_dump_json(indent=2)
    settings_path.write_text(settings_values)


def clear_settings():
    logger.debug(f"Removing saved settings at {settings_path}")
    settings_path.unlink(missing_ok=True)


cli = typer.Typer(help="Configure the app, change settings, or view how it's currently configured")


@cli.command()
@handle_abort
@init_cache
def bind(
    openai_api_key: Annotated[str, typer.Option(help="The API key needed to access OpenAI")],
    resume_path: Annotated[Path, typer.Option(help="The path to your resume")],
    candidate_name: Annotated[
        str,
        typer.Option(help="The name of the candidate to use in the closing"),
    ],
    filename_prefix: Annotated[
        str,
        typer.Option(help="The filename prefix to use for your cover letter"),
    ] = "cover-letter",
    sig_path: Annotated[
        Path | None,
        typer.Option(help="The path to your signature"),
    ] = None,
    heading_path: Annotated[
        Path | None,
        typer.Option(help="The path to your markdown heading"),
    ] = None,
    output_directory: Annotated[
        Path | None,
        typer.Option(
            help="""
                The directory where cover letters should be saved.
                If unset, letters will be saved in the current directory
            """
        ),
    ] = None,
    dev_prompt_path: Annotated[
        Path | None,
        typer.Option(help="An optional path to the developer prompt for letter generation"),
    ] = None,
    user_prompt_template_path: Annotated[
        Path | None,
        typer.Option(help="An optional path to the user prompt template for letter generation"),
    ] = None,
    editor_command: Annotated[
        str | None,
        typer.Option(
            help="""
                Provide a system command to use for opening files in an editor.
                If not provided, will attempt to load system default editor (from env var $EDITOR).
            """
        ),
    ] = None,
):
    """
    Bind the configuration to the app.
    """
    logger.debug(f"Initializing settings with {locals()}")
    settings = init_settings(
        openai_api_key=openai_api_key,
        resume_path=resume_path,
        candidate_name=candidate_name,
        filename_prefix=filename_prefix,
        sig_path=sig_path,
        heading_path=heading_path,
        output_directory=output_directory,
        dev_prompt_path=dev_prompt_path,
        user_prompt_path=user_prompt_template_path,
        editor_command=editor_command,
    )
    dump_settings(settings)


@cli.command()
@handle_abort
@init_cache
@attach_settings(validate=False)
def update(
    ctx: typer.Context,
    openai_api_key: Annotated[str | None, typer.Option(help="The API key needed to access OpenAI")] = None,
    resume_path: Annotated[Path | None, typer.Option(help="The path to your resume")] = None,
    candidate_name: Annotated[str | None, typer.Option(help="The name of the candidate to use in the closing")] = None,
    filename_prefix: Annotated[
        str | None,
        typer.Option(help="The filename prefix to use for your cover letter"),
    ] = None,
    sig_path: Annotated[Path | None, typer.Option(help="The path to your signature")] = None,
    heading_path: Annotated[Path | None, typer.Option(help="The path to your markdown heading")] = None,
    output_directory: Annotated[
        Path | None,
        typer.Option(
            help="""
                The directory where cover letters should be saved.
                If unset, letters will be saved in the current directory
            """
        ),
    ] = None,
    dev_prompt_path: Annotated[
        Path | None,
        typer.Option(help="An optional path to the developer prompt for letter generation"),
    ] = None,
    user_prompt_template_path: Annotated[
        Path | None,
        typer.Option(help="An optional path to the user prompt template for letter generation"),
    ] = None,
    editor_command: Annotated[
        str | None,
        typer.Option(
            help="""
                Provide a system command to use for opening files in an editor.
                If not provided, will attempt to load system default editor (from env var $EDITOR).
            """
        ),
    ] = None,
):
    """
    Update one or more configuration settings that are bound to the app.
    """
    logger.debug(f"Updating settings with {locals()}")
    kwargs: dict[str, Any] = {k: v for (k, v) in locals().items() if v is not None}
    settings = update_settings(ctx.obj.settings, **kwargs)
    dump_settings(settings)


@cli.command()
@handle_abort
@init_cache
@attach_settings(validate=False)
def unset(
    ctx: typer.Context,
    openai_api_key: Annotated[bool, typer.Option(help="The API key needed to access OpenAI")] = False,
    resume_path: Annotated[bool, typer.Option(help="The path to your resume")] = False,
    candidate_name: Annotated[bool, typer.Option(help="The name of the candidate to use in the closing")] = False,
    filename_prefix: Annotated[bool, typer.Option(help="The filename prefix to use for your cover letter")] = False,
    sig_path: Annotated[bool, typer.Option(help="The path to your signature")] = False,
    heading_path: Annotated[bool, typer.Option(help="The path to your markdown heading")] = False,
    output_directory: Annotated[
        bool,
        typer.Option(
            help="""
                The directory where cover letters should be saved.
                If unset, letters will be saved in the current directory
            """
        ),
    ] = False,
    dev_prompt_path: Annotated[
        bool,
        typer.Option(help="An optional path to the developer prompt for letter generation"),
    ] = False,
    user_prompt_template_path: Annotated[
        bool,
        typer.Option(help="An optional path to the user prompt template for letter generation"),
    ] = False,
    editor_command: Annotated[
        bool,
        typer.Option(
            help="""
                Provide a system command to use for opening files in an editor.
                If not provided, will attempt to load system default editor (from env var $EDITOR).
            """
        ),
    ] = False,
):
    """
    Remove a configuration setting that was previously bound to the app.
    """
    logger.debug(f"Updating settings with {locals()}")
    keys = [k for k in locals() if locals()[k]]
    settings = unset_settings(ctx.obj.settings, *keys)
    dump_settings(settings)


@cli.command()
@handle_abort
@init_cache
@attach_settings(validate=False)
def show(ctx: typer.Context):
    """
    Show the config that is currently bound to the app.
    """
    parts = []
    for field_name, field_value in ctx.obj.settings:
        if field_name == "invalid_warning":
            continue
        parts.append((dasherize(field_name), field_value))
    max_field_len = max(len(field_name) for field_name, _ in parts)
    message = "\n".join(f"[bold]{k:<{max_field_len}}[/bold] -> {v}" for k, v in parts)
    if ctx.obj.settings.invalid_warning:
        message += f"\n\n[red]Configuration is invalid: {ctx.obj.settings.invalid_warning}[/red]"
    terminal_message(message, subject="Current Configuration")


@cli.command()
@handle_abort
@init_cache
def clear():
    """
    Clear the config from the app.
    """
    logger.debug("Clearing settings")
    clear_settings()



from datetime import datetime
from dataclasses import dataclass, field
from pathlib import Path

from smart_letters.config import Settings


@dataclass
class CliContext:
    settings: Settings | None = None
    timestamp: str = field(default_factory=lambda: datetime.now().strftime("%Y%m%d.%H%M%S"))


@dataclass
class PromptConfig:
    dev_prompt_path: Path | None
    user_prompt_template_path: Path | None


@dataclass
class Prompts:
    dev_prompt: str
    user_prompt_tempate: str


@dataclass
class LetterConfig:
    resume_path: Path
    candidate_name: str
    filename_prefix: str
    openai_api_key: str
    posting_url: str
    fake: bool
    cache_path: Path
    editor_command: str
    sig_path: Path | None = None
    heading_path: Path | None = None
    example_path: Path | None = None
    output_directory: Path | None = None
    company: str | None = None
    position: str | None = None


@dataclass
class RenderConfig:
    file_stem: str
    timestamp: str
    output_directory: Path | None = None



from typing import Any

from loguru import logger
from mako.template import Template

from smart_letters.exceptions import Abort
from smart_letters.schemas import PromptConfig, Prompts
from smart_letters.utilities import asset_path


def get_prompts(prompt_config: PromptConfig) -> Prompts:
    dev_prompt_path = prompt_config.dev_prompt_path or asset_path("dev_prompt.txt")
    logger.debug(f"Loading dev prompt from {dev_prompt_path}")
    user_prompt_template_path = prompt_config.user_prompt_template_path or asset_path("user_prompt.txt.mako")
    logger.debug(f"Loading user prompt template from {user_prompt_template_path}")

    return Prompts(
        dev_prompt=dev_prompt_path.read_text(),
        user_prompt_tempate=user_prompt_template_path.read_text(),
    )


def build_prompts(prompts: Prompts, **user_prompt_vars) -> list[dict[str, Any]]:
    logger.debug("Assembling prompts for OpenAI")
    try:
        rendered_user_prompt = Template(prompts.user_prompt_tempate).render(**user_prompt_vars)
    except Exception:
        raise Abort("Couldn't render the user prompt", subject="Render failed")

    return [
        dict(role="developer", content=prompts.dev_prompt),
        dict(role="user", content=rendered_user_prompt),
    ]
